package tools.descartes.prisma.agent.generic;

import java.io.File;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.equinox.app.IApplication;
import org.eclipse.equinox.app.IApplicationContext;

import edu.kit.ipd.descartes.identifier.Identifier;
import edu.kit.ipd.descartes.mm.resourcetype.ResourceType;
import edu.kit.ipd.descartes.mm.resourcetype.ResourceTypeRepository;
import edu.kit.ipd.descartes.mm.resourcetype.ResourcetypePackage;
import tools.descartes.prisma.agent.generic.genericagent.GenericAgentConfiguration;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.adapter.ModelSkeletonAdapter;
import tools.descartes.prisma.core.agent.AgentController;
import tools.descartes.prisma.core.agent.DelegationProvider;
import tools.descartes.prisma.core.impl.CDOModelRepository;
import tools.descartes.prisma.model.scope.DelegatedObject;
import tools.descartes.prisma.model.scope.ScopeConfiguration;
import tools.descartes.prisma.model.skeleton.ModelSkeleton;
import tools.descartes.prisma.model.skeleton.SkeletonFactory;

public class GenericAgent implements IApplication {

	private class GenericAgentController extends AgentController {

		private ModelSkeleton skeleton;

		public GenericAgentController(ModelRepository repository, GenericAgentConfiguration configuration) {
			super(repository, configuration.getScope().getPath(), configuration.getId());
		}
	}

	private class GenericDelegationProvider implements DelegationProvider {

		private final Map<String, Map<String, String>> elements = new HashMap<>();

		public GenericDelegationProvider(ScopeConfiguration configuration) {
			for (DelegatedObject curElement : configuration.getDelegatedObjects()) {
				if (curElement.getObject() instanceof Identifier) {
					elements.put(((Identifier) curElement).getId(), curElement.getAdditionalProperties().map());
				}
			}
		}

		@Override
		public boolean needsDelegation(EObject object) {
			if (object instanceof Identifier) {
				return elements.containsKey(((Identifier) object).getId());
			}
			return false;
		}

		@Override
		public Map<String, String> getAdditionalProperties(EObject object) {
			if (object instanceof Identifier) {
				Map<String, String> properties = elements.get(((Identifier) object).getId());
				if (properties != null) {
					return properties;
				}
			}
			return Collections.emptyMap();
		}

	}

	@Override
	public Object start(IApplicationContext context) throws Exception {
		String[] args = (String[]) context.getArguments().get(IApplicationContext.APPLICATION_ARGS);

		Options options = new Options();
		options.addOption("r", "repository", true, "Host of the model repository");
		options.addOption("p", "port", true, "Port of the model repository");
		options.addOption("c", "config", true, "Agent configuration");

		CommandLineParser cliParser = new GnuParser();
		CommandLine cli = cliParser.parse(options, args);

		String repositoryURL = null;
		String configurationPath = null;
		boolean help = false;
		if (cli.hasOption("r")) {
			repositoryURL = cli.getOptionValue("r");
		} else {
			help = true;
		}
		if (cli.hasOption("c")) {
			configurationPath = cli.getOptionValue("c");
		} else {
			help = true;
		}

		if (help) {
			HelpFormatter helpFormat = new HelpFormatter();
			helpFormat.printHelp("genericagent", options);
			return null;
		}
		GenericAgentConfiguration configuration = loadConfiguration(configurationPath);

		ModelRepository repository = new CDOModelRepository(repositoryURL);
		GenericAgentController agent = new GenericAgentController(repository, configuration);
		agent.start().get();

		replacePublicResources(agent, configuration.getSkeleton());

		agent.apply(configuration.getSkeleton());
		
		agent.stop().get();

		return null;
	}

	@Override
	public void stop() {
		// TODO Auto-generated method stub

	}

	private void loadTemplates(GenericAgentController agent, String[] files, String delegationModel) {
		ModelSkeleton skeleton = SkeletonFactory.eINSTANCE.createModelSkeleton();

		// Set<EObject> roots = new HashSet<>();
		// ResourceSet set = new ResourceSetImpl();
		// // IMPORTANT: Always provide full paths to EMF so that
		// cross-references
		// // are resolved correctly.
		// for (String currentFile : files) {
		// Resource res = set.getResource(URI.createFileURI(new
		// File(currentFile).getAbsolutePath()), true);
		// roots.add(res.getContents().get(0));
		// }
		//
		// // Initialize the delegation settings of the agent
		// if (delegationModel != null) {
		// Resource res = set.getResource(URI.createFileURI(new
		// File(delegationModel).getAbsolutePath()), true);
		// final DelegatedElementRepository delegationRepository =
		// (DelegatedElementRepository) res.getContents()
		// .get(0);
		// DelegationProvider provider = new
		// GenericDelegationProvider(delegationRepository);
		// agent.setDelegationProvider(provider);
		// }
		//
		// // IMPORTANT: Resolve all proxy references
		// EcoreUtil.resolveAll(set);
		// Map<EObject, Collection<Setting>> unresolved =
		// EcoreUtil.UnresolvedProxyCrossReferencer.find(set);
		// for (EObject problemObject : unresolved.keySet()) {
		// Collection<Setting> unresolvedSettings =
		// unresolved.get(problemObject);
		// for (Setting curSetting : unresolvedSettings) {
		// System.out.println("Unresolved proxy: " +
		// curSetting.getEStructuralFeature());
		// }
		// }
		//
		// // Make a copy of all loaded elements
		// EcoreUtil.Copier copier = new EcoreUtil.Copier();
		// Collection<EObject> copiedRoots = copier.copyAll(roots);
		// copier.copyReferences();
		//
		// // Now integrate the copies into the model skeleton
		// for (EObject curRoot : copiedRoots) {
		// integrateIntoSkeleton(skeleton, curRoot);
		// }
		//
		// replacePublicResources(agent, skeleton);
		//
		// // Store skeleton into repository
		// agent.apply(skeleton);
	}

	private GenericAgentConfiguration loadConfiguration(String configurationPath) {
		ResourceSet set = new ResourceSetImpl();
		Resource res = set.getResource(URI.createFileURI(new File(configurationPath).getAbsolutePath()), true);
		GenericAgentConfiguration configuration = (GenericAgentConfiguration) res.getContents().get(0);
		return configuration;
	}

	private void integrateIntoSkeleton(ModelSkeleton skeleton, EObject root) {
		for (EReference curReference : skeleton.eClass().getEAllContainments()) {
			if (curReference.getEReferenceType().isSuperTypeOf(root.eClass())) {
				if (curReference.isMany()) {
					@SuppressWarnings("unchecked")
					List<EObject> values = (List<EObject>) skeleton.eGet(curReference);
					values.add(root);
				} else {
					skeleton.eSet(curReference, root);
				}
				return;
			}
		}
	}

	private void replacePublicResources(GenericAgentController agent, ModelSkeleton skeleton) {
		// Resource types (e.g., CPU, HDD, DELAY) which are defined globally are
		// not included in the model skeleton.
		// We need to find these references in the model skeleton and connect
		// them with the instances in the repository.
		TreeIterator<EObject> iterator = skeleton.eAllContents();
		ResourceTypeRepository publicResourceTypes = agent.getContainerScope().getSystemScope().getResourceTypes();
		while (iterator.hasNext()) {
			EObject curObject = iterator.next();
			for (EReference curReference : curObject.eClass().getEAllReferences()) {
				if (!curReference.isContainment() && !curReference.isContainer()) {
					if (ResourcetypePackage.Literals.RESOURCE_TYPE.isSuperTypeOf(curReference.getEReferenceType())) {
						ResourceType curType = (ResourceType) curObject.eGet(curReference);
						// find matching resources in system global scope
						// (matching is based on name and class
						for (ResourceType curPublicOne : publicResourceTypes.getResourceTypes()) {
							if (curType.eClass().isInstance(curPublicOne)
									&& curType.getName().equals(curPublicOne.getName())) {
								ModelSkeletonAdapter adapter = agent.getModelRepository().adapt(curType,
										ModelSkeletonAdapter.class);
								adapter.setRepositoryObject(curPublicOne);
								break;
							}
						}
					}
				}
			}
		}
	}

}
